import java.net.*;
import java.io.*;
import java.rmi.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.awt.*;
import Serveur.*;
import Shared.*;

// Classe d'exception
class Usage extends Exception
{
  Usage()
  {
    super("ImageServeur DATA_DIR\n");
  }
}

/**<dl><dt>
 * Module ID: ImageServeur.java                                                        <dt>
 * Date     : 27 Mars 1997                                                             <dt>
 * Note     : Ce module necessite le JDK 1.1.1                                         <dt>
 *                                                                                     <dt>
 * Revision :                                                                          <dt>
 *                                                                                     <dt>
 *   001      15 Avril 1997 - Olivier Langlois                                         <dt>
 *   A ajouter un SOCKET TIMEOUT au serverSocket afin que le programme puisse terminer <dt>
 *   normalement apres que le flag listening ait change d'etat meme si aucun autre     <dt>
 *   client sollicite le serveur.                                                      <dt>
 *   </dl>
 * @author    Olivier Langlois
 * @version   1.001
 */
public class ImageServeur extends Frame implements Runnable
{

  static final int portNumber = 10012;      //numero du port

  public  TextArea     display;

  private Hashtable    ImageTable;
  private ServerSocket serverSocket;
  private ThreadGroup  Connections;
  private boolean      listening;

  // Constructeur
  /**
   * Initialise le serveur d'image
   * @param f1 repertoire ou se trouve les fichiers de donnees.
   */
  public ImageServeur(File f1) throws IOException, RemoteException
  {
    // On prepare le terminal
    super("Serveur d'images");
    initTerminal();
   
    // Initialise la table d'images.
    initImageTable(f1);
 
    // Initialise la connection serveur
    initConnection();

    // Initialise le flag listening
    listening = true;

    // Initialise le thread 'listener'
    Connections = new ThreadGroup("Connections");
    new Thread(this, "listener").start();
  }

  // Destructeur
  protected void finalize() throws IOException
  {
    // Ferme le socket
    serverSocket.close();
  }

  // Methodes 'envelloppes'
  public boolean isListening()
  {
    return listening;
  }

  public ImageHeader getImageHeader( String keyStr )
  {
    return (ImageHeader)ImageTable.get( keyStr );
  }

  public Enumeration getImageTableKeys()
  {
    return ImageTable.keys();
  }

  public int getImageTableSize()
  {
    return ImageTable.size();
  }

/******************************************************************************
 *
 *             Initialise la fenetre terminal.
 *
 ****************************************************************************/
  private void initTerminal()
  {
    addWindowListener( new ServeurHandler() );
    display = new TextArea( 20, 10 );
    add( display, "Center" );
    setSize( 400, 300 );
    show();
  }

/******************************************************************************
 *
 *             Initialise la table d'images.
 *
 * @param      f1 (File) Repertoire des donnees
 *
 * @exception java.rmi.RemoteException
 *
 ****************************************************************************/
  private void initImageTable(File f1) throws IOException, RemoteException
  {
    ImageTable          = new Hashtable();
    String interFiles[] = f1.list( new OnlyExt("hdr") );

    for( int i = 0; i < interFiles.length; i++ )
    {
      display.append( "Extraction de l'information pertinente du fichier " +
                      interFiles[i] + "\n" );
      try
      {
        ImageHeader parser = new ImageHeader( new
                             FileInputStream(ImageHeader.DATA_DIR + interFiles[i]) );
        switch( parser.readInterfile() )
        {
          case -1:
            display.append( parser.explainError() + "\n" );
            break;
          case 1:
            if( !ImageTable.containsKey( parser.getDataFileName() ) )
            {
              // Bind l'objet ImageHeader au RMI registry.
              display.append( "Bind cet ImageHeader au RMI registry\n" );
              Naming.rebind("//" + InetAddress.getLocalHost().getHostName() + "/" +
                            parser.getDataFileName(), parser);
              // Ajoute le parser a la table d'images.
              ImageTable.put( parser.getDataFileName(), parser );
            }
            break;
        }
      }
      catch (ParseError x)
      {
        display.append( "Erreur de syntaxe : " + x.getMessage() + "\n" );
      }
      catch (FileNotFoundException fe)
      {
        display.append( "Fichier non trouve: " + fe.getMessage() + "\n" );
      }
    }    
  }

/******************************************************************************
 *
 *             Initialise la connection avec le serveur.
 *
 ****************************************************************************/
  private void initConnection()
  {
    // Initialise le socket de type serveur
    try
    {
      serverSocket = new ServerSocket(portNumber);
      serverSocket.setSoTimeout(30000);
    }
    catch (IOException e)
    {
      System.err.println("Ne peut ecouter le port: " + portNumber + ", " + e.getMessage());
      System.exit(1);
    }
  }

/******************************************************************************
 *
 *             Attendre les requetes de clients.
 *
 * Note      : L'appel de cette methode est bloque jusqu'a ce que l'usager
 *             ferme la fenetre.
 *
 ****************************************************************************/
  public void run()
  {
    ImageServeurThread.Serveur = this;
    display.append("Pret a recevoir des clients\n");

    while(listening)
    {
      Socket clientSocket = null;

      try
      {
        clientSocket = serverSocket.accept();
      }
      catch (IOException e)
      {
        if(listening)
          display.append(Thread.currentThread().getName()+"\n"+"Exception par accept(): "+
                         portNumber + ", " + e.getMessage() + "\n");
        continue;
      }
      new ImageServeurThread(clientSocket, Connections).start();
    }
    if( Connections.activeCount() == 0 )
      termineProgramme();
  }

/******************************************************************************
 *
 *             Termine l'ecoute du serverSocket.
 *
 ****************************************************************************/
  public void termineEcoute()
  {
    listening = false;
    dispose();
  }

/******************************************************************************
 *
 *             Termine le programme lorsque l'ecoute est terminee.
 *
 ****************************************************************************/
  public void termineProgramme()
  {
    if(!listening)
      System.exit(0);
  }

  public static void main(String[] args) throws Usage, IOException, RemoteException
  {
    if( args.length != 1 )
      throw new Usage();

    File f1 = new File(args[0]);

    if( !f1.isDirectory() )
      throw new Usage();
    else
    {
      ImageHeader.DATA_DIR = args[0] + "/";
      /*
       * Installe un nouveau gestionnaire de securite pour
       * limiter l'acces aux objets de type ImageHeader.
       */
      System.setSecurityManager(new RMISecurityManager());
      ImageServeur Serveur = new ImageServeur(f1);
    }
  }
}

