package ImageTools;

import java.awt.image.*;
import java.util.*;
import java.io.*;
import java.rmi.RemoteException;
import Shared.*;

/**<dl><dt>
 * Module ID: KodakProducer.java                                                         <dt>
 * Date     : 14 Avril 1997                                                              <dt>
 *                                                                                       <dt>
 * Revision :                                                                            <dt>
 *                                                                                       <dt>
 *   001      15 Avril 1997 - Olivier Langlois                                           <dt>
 *   La methode run() a ete modifie afin qu'elle transmette une serie de frames avec     <dt>
 *   une resolution d'image croissante a chaque nouvelle frame.                          <dt>
 *                                                                                       <dt>
 *   002      16 Avril 1997 - Olivier Langlois                                           <dt>
 *   La methode f_calculPalette() a ete modifie afin qu'elle cree une palette de couleur.<dt>
 *                                                                                       <dt>
 *   003      16 Avril 1997 - Olivier Langlois                                           <dt>
 *   La methode startMatrixProduction() a ete remplace par setMatrix() et la methode     <dt>
 *   startProduction() a ete modifie afin de permettre a cette classe de fonctionner     <dt>
 *   normalement avec la classe java.awt.image.FilteredImageSource.                      <dt>
 *   La variable _t a ete ajoute afin de permettre a la methode startProduction()        <dt>
 *   de verifier si une production est deja en cours avant d'en demarrer une autre.      <dt>
 *                                                                                       <dt>
 *   004      18 Avril 1997 - Olivier Langlois                                           <dt>
 *   Derive la classe KodakProducer de java.util.Observable afin de donner les moyens    <dt>
 *   aux utilisateurs de cette classe de determiner lorsque la production est terminee.  <dt>
 *                                                                                       <dt>
 *   005      18 Avril 1197 - Olivier Langlois                                           <dt>
 *   Ajoute la propriete PIX_MAX_VAL a la table des proprietes pour le filtre            <dt>
 *   SeuilFilter.
 *   </dl>
 * @author    Olivier Langlois
 * @version   1.005
 */
public class KodakProducer extends Observable implements ImageProducer, Runnable,
                                                         KodakFltConst
{
  private Vector            theConsumers;
  private Hashtable         props;
  private RemoteImageHeader Header;
  private Thread            _t;
  private int[]             imageArray;
  private int               matrixNum;
  private int               numImages;
  private int width, height, row, col, ImageOffset;

  /**
   * Initialise un nouvel KodakProducer
   * @param Header Objet contenant l'information relative a l'image
   * @param is     InputStream provenant de la connection avec le serveur d'images
   */
  public KodakProducer( RemoteImageHeader Header, InputStream is ) throws IOException,
                                                                          RemoteException
  {
    theConsumers = new Vector();
    props        = new Hashtable();
    this.Header  = Header;
    numImages    = Header.getnumImages();
    _t           = null;
    matrixNum    = Integer.MAX_VALUE;  // MAX_VALUE == Toutes les matrices de l'image.

    imageArray   = new int[Header.getnumImages() * Header.getmatrixSize_X() *
                           Header.getmatrixSize_Y()];
    f_receiveData(is);
  }

  private void f_receiveData( InputStream is ) throws IOException, RemoteException
  {
    byte val[] = new byte[2];
    boolean EndianFlag = Header.getEndianFlag();

    for (int offset = 0; offset < imageArray.length; offset++)
    {
      is.read(val, 0, val.length);
      if (EndianFlag)
        imageArray[offset] = val[1] << 8 | val[0] & 255;
      else
        imageArray[offset] = val[0] << 8 | val[1] & 255;
    }
  }

  public void addConsumer(ImageConsumer ic)
  {
    if (isConsumer(ic) == false) theConsumers.addElement(ic);
  }

  public void removeConsumer(ImageConsumer ic)
  {
    if (isConsumer(ic) == true) theConsumers.removeElement(ic);
  }

  public boolean isConsumer(ImageConsumer ic)
  {
    return theConsumers.indexOf(ic) > -1;
  }

/******************************************************************************
 *
 *             Produit l'image constituee de toutes les matrices disponibles.
 *
 * Note      : Cette methode ne doit pas etre reappeler avant que
 *             tous les consommateurs aient recu ImageConsumer.IMAGEERROR
 *             ou ImageConsumer.STATICIMAGEDONE.
 *
 * @param      ic (ImageConsumer) Consommateur a ajouter dans la liste.
 * @see        java.awt.image.ImageProducter#startProduction
 *
 ****************************************************************************/
  public void startProduction(ImageConsumer ic)
  {
    addConsumer(ic);

    if( _t == null )
    {
      _t = new Thread(this, "startProduction thread");

      try
      {
        if( matrixNum == Integer.MAX_VALUE )
        {
          // Determine la disposition des images.
          if( (numImages % 4) == 0 )
          {
            if( numImages == 4 )
            {
              row = col = 2;
            }
            else
            {
              col = 4;
              row = numImages / 4;
            }
          }
          else if( (numImages % 3) == 0 )
          {
            col = 3;
            row = numImages / 3;
          }
          else
          {
            col = numImages;
            row = 1;
          }

          width  = col * Header.getmatrixSize_X();
          height = row * Header.getmatrixSize_Y();
          ImageOffset = 0;
        }  // Fin de IF matrixNum == MAX_VALUE
        else
        {
          width  = Header.getmatrixSize_X();
          height = Header.getmatrixSize_Y();
          ImageOffset = width * height * matrixNum;
          row = col = 1;
        }
        _t.start();
      }
      catch( RemoteException re )
      {
        System.out.println( re.getMessage() );

        // Avertit tous les consommateurs que l'operation a echoue.
        ImageConsumer currentIC;
        Vector v = (Vector)theConsumers.clone();
        Enumeration e = v.elements();

        while (e.hasMoreElements())
        {
          currentIC = (ImageConsumer)e.nextElement();
          currentIC.imageComplete(ImageConsumer.IMAGEERROR);
        }
        _t = null;
        setChanged();
        notifyObservers();
      }
    }  // Fin de IF _t == null
  }

/******************************************************************************
 *
 *             Indique a KodakProducer qu'elle matrice produire.
 *
 * @param matrixNum (int) # de la matrice (ou 0 est la 1iere matrice et
 *                                         Integer.MAX_VALUE == toutes les matrices).
 * @exception KodakException
 *
 ****************************************************************************/
  public void setMatrix( int matrixNum ) throws KodakException
  {
    if( matrixNum >= numImages )
      throw new KodakException("matrixNum invalide : matrixNum " + matrixNum +
                               " numImages " + numImages );
    else this.matrixNum = matrixNum;
  }

  public void run()
  {
    ImageConsumer ic;
    Vector v = (Vector)theConsumers.clone();
    Enumeration e = v.elements();

    int matrixSize_X = width  / col;
    int matrixSize_Y = height / row;
    int matrixSize   = matrixSize_X * matrixSize_Y;
    int _nWt = width  / 16;  // Grandeur apparente
    int _nHt = height / 16;
    int x, y, w, h, offset;
    ColorModel cm = f_calculPalette();

    // Envoie les caracteristiques de l'image aux consommateurs
    for (e = v.elements(); e.hasMoreElements(); )
    {
      ic = (ImageConsumer)e.nextElement();

      ic.setColorModel(cm);
      ic.setDimensions(width, height);
      ic.setProperties(props);
      ic.setHints(ImageConsumer.RANDOMPIXELORDER);
    }

    while (_nWt <= width && _nHt <= height)
    {
      w = width  / _nWt;
      h = height / _nHt;
      int [] imageBlock = new int [w * h];

      // Envoie la valeur de chaque pixel apparent aux consommateurs qui pour eux
      // un pixel consiste en un bloc de pixels selon la resolution apparente.
      for (int nY = 0; nY < _nHt; nY++)
      {
        y = nY * h;

        for (int nX = 0; nX < _nWt; nX++)
        {
          // Ne repete pas le calcul d'un point.
          if (nY % 2 == 0 && nX % 2 == 0 && _nWt != width/16 && _nHt != height/16)
            continue;

          x = nX * w;

          offset = (y/matrixSize_Y)*matrixSize*col + (y%matrixSize_Y)*matrixSize_X +
                   (x/matrixSize_X)*matrixSize     + (x%matrixSize_X) + ImageOffset;

          for (int j = 0; j < w * h; j++) imageBlock[j] = imageArray[offset];

          for (e = v.elements(); e.hasMoreElements(); )
          {
            ic = (ImageConsumer)e.nextElement();

            ic.setPixels(x, y, w, h, cm, imageBlock, 0, w);
          }
        }   // Fin du FOR(nX)
        // Avertit les consommateurs qu'une frame a ete livree.
        for (e = v.elements(); e.hasMoreElements(); )
        {
          ic = (ImageConsumer)e.nextElement();

          ic.imageComplete(ImageConsumer.SINGLEFRAMEDONE);
        }
      }     // Fin du FOR(nY)
      _nWt *= 2;
      _nHt *= 2;
    }  // Fin du WHILE

    // Avertit les consommateurs que la transmission de l'image est terminee.
    for (e = v.elements(); e.hasMoreElements(); )
    {
      ic = (ImageConsumer)e.nextElement();

      ic.imageComplete(ImageConsumer.STATICIMAGEDONE);
    }
    _t = null;
    setChanged();
    notifyObservers();
  }

  private ColorModel f_calculPalette()
  {
    int UpLimit = width * height + ImageOffset;
    int max     = 0;

    for( int i = ImageOffset; i < UpLimit; i++ )
    {
      if( imageArray[i] > max )
        max = imageArray[i];
    }

    props.put( PIX_MAX_VAL, new Integer(max) );

    byte R[] = new byte[max + 1];
    byte G[] = new byte[max + 1];
    byte B[] = new byte[max + 1];
    int interval=max/5;

    for (int i = 0; i < R.length; i++)
    {
      if (i<interval)
      {
        B[i]=(byte) (255*i/interval);
      }
      else if (i<2*interval)
      {
        B[i]=(byte) (255);
        G[i]=(byte) (255*i/(interval*2));
      }
      else if (i<3*interval)
      {
        B[i]=(byte) (255-255*i/(interval*3));
        G[i]=(byte) (255);
      }
      else if (i<4*interval)
      {
        G[i]=(byte) (255);
        R[i]=(byte) (255*i/(interval*4));
      }
      else
      {
        G[i]=(byte) (255-255*i/(interval*5));
        R[i]=(byte) (255);
      }

    }
    return new IndexColorModel(8, max + 1, R, G, B);
  }

  public void requestTopDownLeftRightResend(ImageConsumer ic)
  {
    ;
  }
}

